! Thursday  7th December version
!+
! Module: omsrclistcomb_mod
!
! Authors:  Tim Harries - MSSL
!          
! Description: Combines OSW source list into an observation source list
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!
! History: 
!        98 Feb: created TJH
!        98 Apr 2: changed column names for colours in the table
!                  from eg. B-V to B_V to conform to OGIP FITS
!        98 April 27: changed coord calcs to double (TJH)
!        98 May 14: Added columnExist function in reponse to spr
!        99 May 28: Changed parameter names to conform to new convention.
!        2nd June 00 : Added galactic coordinates as 8 byte data
!        4th June 00 : Modified ra and dec to be 8 byte data
!        4th June 00 : Added memory clean-up at  end  of routine
!        7th July 00 : Ammended code to reflect change of column
!        heading from NOBS to SRC_ID. (SSC-SPR-1911) Low memory failure.
!        5th September 00 : General tidy of the memory allocation and deallocation.	
!        20th September 00 : Replace NULLVALUE by the result of the division 1.0/0.0 
!        This results in the entry NULL in the table which indicates that a ource has
!        not been detected  in a filter.
!        21st September 00 : Added the following columns POSERR,SIGNIFICANCE,FWHM_MAJ & FWHM_MIN
!        21st September 00 : Added in the units for the table
!        27th September 00 : Corrected the call message parameters.
!        2nd November   00 : Range validation of instrumental magnitudes
!        carried out before standard colour calculated using CAL call 
!        24th May 2001 Modified averaging routine so that if any error is zero
!                     the weighting factors are set to 1                
!-
!
module omsrclistcomb_mod
  use param
  use cal
  use dal
  use Errorhandling
  use sla

  implicit none 

  private 

  public omsrclistcomb

  !*****************************
  ! Structure to defines a color
  !*****************************
  type ColourT
     integer :: mag1
     integer :: mag2
     character(len=20) label
     character(len=20) elabel
  end type ColourT

contains 

  !---------------------------------------------------------------------
  subroutine omsrclistcomb
  !---------------------------------------------------------------------

    integer, parameter               :: nFilt = 7             ! no of filters
    integer, parameter               :: nCol = 15             ! no of colours for instrumental magnitudes
    integer, parameter               :: nCol1 = 3             ! no of colours for standard magnitudes
    type(ColourT)                    :: Colour(nCol)
    type(DataSetT)                   :: oswDataSet            ! handles
    type(DataSetT)                   :: outDataSet            ! handle for output dataset
    type(TableT)                     :: tableSet
    type(ColumnT) :: pColumn
  
    character(len=10),allocatable    :: filter(:)              ! filters for osw source lists
    character(len=5)                 :: filtid(nFilt)
    character(len=1024), dimension(:), allocatable    :: oswFile        ! filenames
    character(len=50)                :: label                  ! column label
    integer                          :: nmag                   ! no of mags
    integer                          :: nStart                 ! start no of sources
    integer                          :: nFound                 ! no of identical sources found
    integer, allocatable             :: iIndep(:)
    integer                          :: nIndep                 ! number of independent sources
    integer, allocatable             :: list(:)                ! list of identical sources
    integer, allocatable             :: oswFound(:)
    integer                          :: nosw                   ! no of osw lists
    integer                          :: i,j,k                  ! counters
    integer                          :: nSources               ! total no of sources
    integer,allocatable              :: nrows(:)               ! no of rows in each osw source list
    integer                          :: ntemp                  ! temp storage
    integer, allocatable             :: oswtoobs(:,:)          ! cross-links
    integer, allocatable             :: oswid(:)               ! no of each osw
    integer(kind=int8), allocatable  :: qflag(:)      
    integer(kind=int8), allocatable  :: cflag(:)      
    integer(kind=int8), allocatable  :: eflag(:)      
    integer(kind=int32), pointer :: sourceID(:)
    integer :: ndims(1) = (/1/)
    real(kind=double)                :: ra_tmp,dec_tmp         ! tmp storage for coords
    real(kind=double), parameter     :: pi = 3.14159265358979323846d0      ! PI
    real(kind=double), parameter      :: degreesToRadians = pi / 180.0d0
    real(kind=double), parameter      :: RadiansToDegrees = 180.0d0 / pi
    real(kind=single), allocatable   :: mag_table(:,:)         ! table of magnitudes
    real(kind=single), allocatable   :: magerr_table(:,:)      ! table of magnitude errors
    real(kind=double), allocatable   :: distlist(:)            ! list of distances
    real(kind=double)                :: dist                   ! distance between two sources
    real(kind=double)                :: err                    ! error on that distance
    real(kind=single)                :: NullValue              ! Null
    real(kind=single)                :: nsig                   ! tolerance for source id
    real(kind=double), allocatable   :: ra(:)                  ! source list columns
    real(kind=double), allocatable   :: dec(:)                   
    real(kind=single), allocatable   :: mag(:)        
    real(kind=single), allocatable   :: magerr(:)       
    real(kind=double)                :: weight
    real(kind=double)                :: weightSum
    real(kind=single), allocatable   :: poserr(:)   
    logical, allocatable             :: magexist(:,:)            ! does a magnitude exist?
    logical, dimension(1 : nCol)     :: stdMagExist              ! Does a standard magnitude exist?
    real(kind=single), allocatable   :: magColour(:)   
    real(kind=single), allocatable   :: magColourError(:)   
    real(kind=single), allocatable   :: standardMagnitude(:, :)   
    real(kind=single), allocatable   :: standardMagnitudeError(:, :)   
 
    integer                          :: allocErr
    integer(kind=int32)              :: filt(7)                  ! Filter magnitude.
    real(kind=single), allocatable   :: significance(:)          ! source list columns
    real(kind=single), allocatable   :: significanceTable(:,:)   ! table of magnitudes
    real(kind=single)                :: significanceTemp         ! Temporary storage
    real(kind=single), allocatable   :: fullHalfWidthMajor(:)    ! source list columns
    real(kind=single), allocatable   :: majorTable(:,:)          ! table of magnitudes
    real(kind=single)                :: majorTemp                ! Temporary storage
    real(kind=single), allocatable   :: fullHalfWidthMinor(:)    ! source list columns
    real(kind=single), allocatable   :: minorTable(:,:)          ! table of magnitudes
    real(kind=single)                :: minorTemp                ! Temporary storage
    real(kind=single)                :: poserr_tmp
    real(kind=double), allocatable   :: galong(:)                ! array for galactic longitudes
    real(kind=double), allocatable   :: galat(:)                 ! array for galactic latitudes    
    real(kind=double)                :: stdColour
    real(kind=double)                :: stdMagnitude1
    real(kind=double)                :: stdMagnitude2
    real(kind=double)                :: stdColourError
    real(kind=double)                :: stdMagnitude1Error
    real(kind=double)                :: stdMagnitude2Error
    real(kind=double), dimension(:), allocatable :: flux, fluxError
    real(kind=double)                :: mean
    real(kind=double)                :: standardDEviation    
    integer            :: noColours
    character(len=100), dimension(15) :: colourString
    character(len=100), dimension(15) :: colourErrorString
    real(kind=single), allocatable, dimension(:, :) :: stdColourMagnitude
    real(kind=single), allocatable, dimension(:, :) :: stdColourMagnitudeError
    integer, allocatable, dimension(:) :: RASortIndex
    integer iColumn, jColumn      
    integer :: noFilters
    integer :: filterIndex(20)    
    integer, dimension(:), allocatable :: index
    data filtid /'WHITE','UVW2','UVM2','UVW1','U','B','V'/ 
    data filt / WHITE,UVW2,UVM2,UVW1,U,B,V /

    !************************
    ! Iniitialise NullValue
    !************************
    NullValue = -real(99999)
  
    !************************************************
    ! Initialise the filter prority and colour tables.
    !************************************************
    call SetUpColourTable(Colour)

    !************************************************
    ! Get the number of osw files and their names.
    !************************************************

     nosw = parameterCount("sourcelistsets")
     allocate(oswfile(1 : nosw), stat=AllocErr)
     call CheckAllocation(allocErr, "oswfile")
 
    do i = 1, nOsw
       oswfile(i) = stringParameter("sourcelistsets", i-1)
    enddo
    
    nsig = realParameter("nsigma")

    !******************************************
    ! create the output observation source list
    !******************************************
    call CreateObsSourceList(nOsw, oswFile, outDataSet)    
 
    allocate(nrows(nosw),stat=allocErr)
    call CheckAllocation(allocErr, "nrows")
 
    allocate(filter(nosw), stat=AllocErr)
    call CheckAllocation(allocErr, "nosw")
 
    !***************************************
    ! figure out how many sources there are
    !***************************************
    nSources = 0 
    do i = 1 , nosw

       oswDataSet = DataSet(oswFile(i),read)
       tableSet = table(oswDataSet,0)
       nrows(i) = numberOfRows(tableSet)
       nSources = nSources + nrows(i)
       call release(tableSet)
       call release(oswDataSet)

    enddo

    !******************************************
    ! Allocate memory for the arrays to hold
    ! the input data.
    !******************************************
    allocate(oswid(1 : nSources), stat=allocErr)
    call CheckAllocation(allocErr, "oswid")

    allocate(ra(1 : nSources), stat=allocErr)
    call CheckAllocation(allocErr, "ra")

    allocate(dec(1 : nSources), stat=allocErr)
    call CheckAllocation(allocErr, "dec")

    allocate(significance(1 : nSources), stat=allocErr)
    call CheckAllocation(allocErr, "significance")
  
    allocate(fullHalfWidthMinor(1 : nSources), stat=allocErr)
    call CheckAllocation(allocErr, "fullHalfWidthMinor")

    allocate(fullHalfWidthMajor(1 : nSources), stat=allocErr)
    call CheckAllocation(allocErr, "fullHalfWidthMajor")
   
    allocate(galong(1 : nSources), stat=allocErr)
    call CheckAllocation(allocErr, "galong")

    allocate(galat(1 : nSources), stat=allocErr)
    call CheckAllocation(allocErr, "galat")

    allocate(poserr(1 : nSources), stat=allocErr)
    call CheckAllocation(allocErr, "poserr")

    allocate(mag(1 : nSources), stat=allocErr)
    call CheckAllocation(allocErr, "mag")
 
    allocate(magerr(1 : nSources), stat=allocErr)
    call CheckAllocation(allocErr, "magerr")
 
    allocate(qflag(1 : nSources), stat=allocErr)
    call CheckAllocation(allocErr, "qflag")

    allocate(cflag(1 : nSources), stat=allocErr)
    call CheckAllocation(allocErr, "cflag")

    allocate(eflag(1 : nSources), stat=allocErr)
    call CheckAllocation(allocErr, "eflag")

    allocate(oswtoobs(1 : nSources, 3), stat=allocErr)
    call CheckAllocation(allocErr, "oswtoobs")

    !********************************************************
    ! Read in the data from each osw source list into the 
    ! appropriate array- appending the new data to the end of
    ! that array.
    !********************************************************
      
    nStart = 0
    do i = 1 , nosw
       oswDataSet = DataSet(oswFile(i),read)
       filter(i) = trim(stringAttribute(oswDataSet, "FILTER"))
       tableSet = table(oswDataSet,0)
 
 

       forall (j = 1  : nrows(i))
          oswid(nStart + j) = i
          oswtoobs(nStart + j, 2) = j
       end forall

       call ReadReal64Column(tableSet, "RA", nrows(i), nStart, ra)
       call ReadReal64Column(tableSet, "DEC", nrows(i), nStart, dec)
       call ReadReal32Column(tableSet, "FWHM_MIN", nrows(i), nStart, fullHalfWidthMinor)
       call ReadReal32Column(tableSet, "FWHM_MAJ", nrows(i), nStart, fullHalfWidthMajor)
       call ReadReal64Column(tableSet, "LII", nrows(i), nStart, galong)
       call ReadReal64Column(tableSet, "BII", nrows(i), nStart, galat)
       call ReadReal32Column(tableSet, "POSERR", nrows(i), nStart, poserr)
       call ReadReal32Column(tableSet, "SIGNIFICANCE", nrows(i), nStart, significance)
       call ReadReal32Column(tableSet, "MAG", nrows(i), nStart, mag)
       call ReadReal32Column(tableSet, "MAGERR", nrows(i), nStart, magerr)
       call ReadInt8Column  (tableSet, "QFLAG", nrows(i), nStart, qflag)
       call ReadInt8Column  (tableSet, "CFLAG", nrows(i), nStart, cflag)
       call ReadInt8Column  (tableSet, "EFLAG", nrows(i), nStart, eflag)

       call release(tableSet)
       call release(oswDataSet)
       nStart = nStart + nrows(i)

    end do


  
    !***************************************
    ! Do a few conversions
    !***************************************
    ra = ra * degreesToRadians
    dec = dec * degreesToRadians
    galong = galong * degreesToRadians
    galat = galat * degreesToRadians

    !***************************************
    ! Allocate memory for output arrays.
    !***************************************
    allocate(list(1 : nSources), stat=allocErr)
    call CheckAllocation(allocErr, "list")

    allocate(oswFound(1 : nSources), stat=allocErr)
    call CheckAllocation(allocErr, "oswFound")
  
    allocate(distlist(1 : nSources), stat=allocErr)
    call CheckAllocation(allocErr, "distlist")
  
    allocate(mag_table(1 : nSources, 1 : nFilt), stat=allocErr)
    call CheckAllocation(allocErr, "mag_table")

    allocate(magerr_table(1 : nSources, 1 : nFilt), stat=allocErr)
    call CheckAllocation(allocErr, "magerr_table")

    allocate(significanceTable(1 : nSources, 1 : nFilt), stat=allocErr)
    call CheckAllocation(allocErr, "significanceTable")
  
    allocate(majorTable(1 : nSources,1 : nFilt), stat=allocErr)
    call CheckAllocation(allocErr, "majorTable")

    allocate(minorTable(1 : nSources, 1 : nFilt),stat=allocErr)
    call CheckAllocation(allocErr, "minorTable")

  
    allocate(magexist(1 : nSources, 1 : nFilt), stat=allocErr)
    call CheckAllocation(allocErr, "magExist")

  
    mag_table = NullValue
    magerr_table = NullValue
    significanceTable = NullValue
    majorTable = NullValue
    minorTable = NullValue
    magexist = .false.

    allocate(flux(1 : nSources), stat=allocErr)
    call CheckAllocation(allocErr, "flux")
    allocate(fluxError(1 : nSources), stat=allocErr)
    call CheckAllocation(allocErr, "fluxError")

    !******************************************************
    ! Find how many identical sources there are in 2 or more
    ! of the osw files. 
    !******************************************************

    do i = 1 , nSources

       nFound = 0

       !**********************************************************
       ! if oswid(n) <= 0 then then we have found the counterparts
       !  of this source- otherwise this source is the first
       !**********************************************************
       if (oswid(i) > 0) then
          nFound = 1
          list(nFound) = i
          distlist(nFound) = 0.0d0
       end if
       !******************************************
       ! now loop over all the rest of the sources
       !******************************************

       do j = i + 1, nSources
 
          !************************************************
          ! make sure this source hasn't been found already 
          !************************************************
          if ((oswid(i) .ne. oswid(j)) .and. (oswid(i) > 0) .and. (oswid(j) > 0)) then

             !*************************************
             ! calculate the distance to the source
             ! and its error (both in arcsecs)
             !*************************************
             dist = ArcLength(ra(i), dec(i), ra(j), dec(j))
             err = AddErrors(dble(poserr(i)), dble(poserr(j)))
          
             !****************************************************** 
             ! if the distance is less than the error then these are 
             ! (might be) the same source
             !******************************************************
             if (dist <  (dble(nsig) * err)) then  ! nsig is a task parameter
                nFound = nFound + 1
                list(nFound) = j
                distlist(nFound) = dist        ! add it to the list
                oswFound(nFound) = oswid(j)
             endif
          endif
       enddo

       j = 1
       do while(j < nFound)
          if (oswFound(j) .eq. oswFound(j + 1)) then
             if (distlist(j + 1) > distlist(j)) then
                if (j + 1 .eq. nFound) then
                   nFound = nFound - 1
                else
                   do k = j + 1, nFound - 2
                      list(k) = list(k+1)
                      distlist(k) = distlist(k + 1)
                      oswFound(k) = oswFound(k + 1)
                   enddo
                   nFound = nFound - 1
                endif
             endif

          endif
          j = j + 1
       end do

       j = 1
       do while(j < nFound)
          if (oswFound(j) .eq. oswFound(j + 1)) then
             if (j + 1 .eq. nFound) then
                nFound = nFound - 1
             else
                do k = j + 1, nFound - 2
                   list(k) = list(k + 1)
                   distlist(k) = distlist(k + 1)
                   oswFound(k) = oswFound(k + 1)
                enddo
                nFound = nFound - 1
             endif
          else
             j = j + 1
          endif
       end do

       if (nFound > 0) then
          write(errstr, *) "Same sources: ", list(1 : nFound)

          call message(AppMsg, Verbose)
          write(errstr, *) "distances: ", real(distlist(1 : nFound))

          call message(AppMsg, Verbose)
 
          !*************************************** 
          ! find the weighted mean of the position
          !***************************************       
          ra_tmp = 0.0d0
          dec_tmp = 0.0d0
          poserr_tmp = 0.0d0
          weightSum = 0.0

          do k = 1 , nFound
             weight = 1.d0 / (poserr(list(k)) * poserr(list(k)))
             weightSum = weightSum + weight
             ra_tmp = ra_tmp + ra(list(k)) * weight
             dec_tmp = dec_tmp + dec(list(k)) * weight
             poserr_tmp = poserr_tmp + poserr(list(k)) * weight
          enddo

          ra(i) = ra_tmp / weightSum
          dec(i) = dec_tmp / weightSum
          poserr(i) = poserr_tmp / weightSum

          !*************************************************** 
          ! Compute the galactic longitude and latitude 
          ! from the averaged RA and dec
          !***************************************************
          call sla_EQGAL(ra(i), dec(i), galong(i), galat(i)) 
          
          !*****************************************************
          ! Go through and fill up the magnitude matrix
          !  
          ! loop over all the filters
          !******************************************************
          do j = 1 , nFilt

             nmag = 0
             significanceTemp = 0.0
             minorTemp = 0.0
             majorTemp = 0.0

             !*********************************
             ! loop over all identical sources
             !*********************************
             do k = 1 , nFound

                if ( trim(filter(oswid(list(k)))) .eq. trim(filtid(j)) ) then

                   !*******************************************************        
                   ! if two sources are in the same filter then
                   ! add it on
                   ! Find the average magnitude by converting the magnitudes
                   ! to fluxes, finding the average flux and then converting 
                   ! back. A weighted average is used.
                   !******************************************************** 
                   if (mag(list(k)) .gt. 0.0) then     
                       nmag = nmag + 1      
                       flux(nmag) = 10.0d0 ** (-0.4d0 * mag(list(k)))                 
                       fluxError(nmag) = magerr(list(k)) * flux(nmag)
                       significanceTemp = significanceTemp + significance(list(k))
                       majorTemp = majorTemp + fullHalfWidthMajor(list(k))
                       minorTemp = minorTemp + fullHalfWidthMinor(list(k))
                   end if
               endif

             end do
              
             !**********************************
             ! Compute the average values
             !**********************************
             if (nmag > 0) then

                significanceTemp = significanceTemp / real(nmag)
                majorTemp = majorTemp / real(nmag)
                minorTemp = minorTemp / real(nmag)

                call ComputeAverage(flux, fluxError, nmag, mean, standardDeviation)        
                mag_table(i, j) = -2.5d0 * dlog10(mean)  
                magerr_table(i, j) = standardDeviation / mean
                significanceTable(i, j) = significanceTemp
                majorTable(i, j) = majorTemp
                minorTable(i, j) = minorTemp
                magexist(i, j) = .true.
             endif

          enddo

          !*************************************************
          ! Need to remember where osw each  source came from
          !*************************************************
          forall (k = 1  : nFound)
             oswtoobs(list(k), 3) = list(1)
             oswid(list(k)) = 0         ! take the source out of the running
          end forall

          oswid(list(1)) = -1     ! we flag the first source found 
       end if

    end do
    deallocate(flux, stat=allocErr)
    call CheckDeallocation(allocErr, "flux")

    deallocate(fluxError, stat=allocErr)
    call CheckDeallocation(allocErr, "fluxError")

    deallocate(list, stat=allocErr)
    call CheckDeallocation(allocErr, "list")

    deallocate(distList, stat=allocErr)
    call CheckDeallocation(allocErr, "distList")
 
    allocate(iIndep(1 : nSources),stat=allocErr)
    call CheckAllocation(allocErr, "iIndep")

    deallocate(oswFound, stat=allocErr)
    call CheckDeallocation(allocErr, "oswFound")

    deallocate(filter, stat=allocErr)
    call CheckDeallocation(allocErr, "filter")

    !*******************************************
    ! Compute the number of independent sources
    ! i.e. those with a oswid(n) = -1
    ! and reorder the arrays.
    !********************************************
    ntemp = 0
    do i = 1 , nSources

       if (oswid(i) < 0) then
          ntemp = ntemp + 1
          ra(ntemp) = ra(i)
          dec(ntemp) = dec(i)
          fullHalfWidthMinor(ntemp) =  fullHalfWidthMinor(i)
          fullHalfWidthMajor(ntemp) =  fullHalfWidthMajor(i)
          galong(ntemp) = galong(i)
          galat(ntemp) = galat(i)
          poserr(ntemp) = poserr(i)
          significance(ntemp) = significance(i)
          qflag(ntemp) = qflag(i)
          cflag(ntemp) = cflag(i)
          eflag(ntemp) = eflag(i)

          do j = 1, nFilt
              mag_table(ntemp, j) = mag_table(i, j)
              magerr_table(ntemp, j) = magerr_table(i, j)
              significanceTable(ntemp, j) = significanceTable(i, j)
              majorTable(ntemp, j) = majorTable(i, j)
              minorTable(ntemp, j) = minorTable(i, j)
              magexist(ntemp, j) = magexist(i, j)
          end do           

       endif
       iIndep(i) = ntemp
    enddo

    nIndep = ntemp

    ! let the user know for info


    write(errstr,'(i6,a)') ntemp,' independent sources'
    call message(AppMsg, Verbose)

    deallocate(oswid, stat=allocErr)
    call CheckDeallocation(allocErr, "oswid")

    deallocate(fullHalfWidthMinor, stat=allocErr)
    call CheckDeallocation(allocErr, "fullHalfWidthMinor")
    
    deallocate(fullHalfWidthMajor, stat=allocErr)
    call CheckDeallocation(allocErr, "fullHalfWidthMajor")
   
    !**************************************************
    ! Create the observation source list
    ! which has nIndep rows
    !**************************************************
    tableSet = addTable(outDataSet, "SRCLIST", nIndep)

    allocate(RASortIndex(1 : nSources), stat=allocErr)
    call CheckAllocation(allocErr, "index")

    forall (i = 1 : nIndep) RASortIndex(i) = i
   
    call AddInt32Column(tableSet, "SRCID", "", nIndep, RASortIndex, RASortIndex)
 
    !*****************************************************
    ! Find the order of increasing  RA- store in RASortIndex
    ! for adding data to source list table in  
    ! increasing RA order (ra is not sorted)
    !****************************************************
    call HeapSort(ra, RASortIndex, nIndep)

    !********************************************
    ! Add the refined ra's and dec's to the table 
    !********************************************
    ra = ra * radiansToDegrees
    call AddReal64Column(tableSet, "RA", "degrees", nIndep, ra, RASortIndex, nullValue)
    deallocate(ra, stat = allocErr)
    call CheckDeallocation(allocErr, "ra")
    dec = dec * radiansToDegrees
    call AddReal64Column(tableSet, "DEC", "degrees", nIndep, dec, RASortIndex, nullValue)
    deallocate(dec, stat = allocErr)
    call CheckDeallocation(allocErr, "dec")
   
    call AddReal32Column(tableSet, "POSERR", "arcsec", nIndep, poserr, RASortIndex, nullValue)
    deallocate(poserr, stat = allocErr)
    call CheckDeallocation(allocErr, "poserr")
    galong = galong * radiansToDegrees
    call AddReal64Column(tableSet, "LII", "degrees", nIndep, galong, RASortIndex, nullValue)
    deallocate(galong, stat = allocErr)
    call CheckDeallocation(allocErr, "galong")
 
    galat = galat * radiansToDegrees
    call AddReal64Column(tableSet, "BII", "degrees", nIndep, galat, RASortIndex, nullValue)
    deallocate(galat, stat = allocErr)
    call CheckDeallocation(allocErr, "galat")
    
    !**************************************
    ! Find out what filters have magnitudes
    !************************************** 
    noFilters = 0
    do i = 1 , nFilt
   
       if (ANY(magexist(1 : nIndep, i : i)))  then
           noFilters = noFilters + 1
           filterIndex(noFilters) = i
       else
        
          write(errstr,*) "No sources found in filter ",filtid(i)
          call message(AppMsg, Verbose)

       end if

    end do
 
    !******************************************************
    ! Add the significance arrays
    !****************************************************** 
    do j = 1 , noFilters
       i = filterIndex(j)
       
       label = trim(filtid(i)) // "_SIGNIF"
       call Add2Real32Column(tableSet, label, "sigma", nIndep, significanceTable, i, RASortIndex, nullValue)
    end do
 
    deallocate(significance, stat=allocErr)
    call CheckDeallocation(allocErr, "significance")

    deallocate(significanceTable, stat=allocErr)
    call CheckDeallocation(allocErr, "significanceTable")
   


    !*********************************************
    ! Add the instrumental magnitude and its error
    !*********************************************
    do j  = 1 , noFilters
        i = filterIndex(j)

        label = trim(filtid(i)) // "_INS"
        call Add2Real32Column(tableSet, label, "magnitudes", nIndep, mag_table, i, RASortIndex, nullValue)

        label = trim(filtid(i)) // "_ERR_INS"
        call Add2Real32Column(tableSet, label, "magnitudes", nIndep, magerr_table, i, RASortIndex, nullValue)
    end do

    !*********************************************
    ! Add the instrumental colours
    !*********************************************
    allocate(magColour(1 : nIndep), stat=allocErr)
    call CheckAllocation(allocErr, "magColour")

    allocate(magColourError(1 : nIndep), stat=allocErr)
    call CheckAllocation(allocErr, "magColourError")
 
    do i = 1, nCol     ! Loop through all the possible colours
       magColour = NullValue
       magColourError = NullValue
       iColumn = Colour(i)%mag1
       jColumn = Colour(i)%mag2

       if (ANY(magexist(1 : nIndep, iColumn : iColumn)) .and.  &
            ANY(magexist(1 : nIndep, jColumn : jColumn))) then

          do k = 1, nIndep
              if (mag_table(k, iColumn) .ne. NullValue .and. &
                  mag_table(k, jColumn) .ne. NullValue) then
                  magColour(k) = mag_table(k, iColumn) - mag_table(k, jColumn)
                  magColourError(k) = AddErrors(dble(magerr_table(k, iColumn)), dble(magerr_table(k, jColumn)))
              end if
          end do 

          label = trim(Colour(i)%label) // "_INS"
          
          call AddReal32Column(tableSet, label, "magnitudes", nIndep, magColour, RASortIndex, nullValue)

          label = trim(Colour(i)%elabel) // "_INS"
          call AddReal32Column(tableSet, label, "magnitudes", nIndep, magColourError, RASortIndex, nullValue)

       end if
    end do

    !*******************************************************************
    ! now we have the colours
    ! we can currently have (U-B) and/or (B-V), but the combinations
    ! can be added straightfowardly by changing the array in SetUpColours
    !********************************************************************
    call CAL_setState(OM , 1, accuracyLevel = ACCURACY_LOW)

    !****************************************
    ! Allocate memory for arrays holding
    ! the standard colours
    !*************************************** 
    allocate(stdColourMagnitude(1 : nIndep, 1 : nFilt), stat=allocErr)
    call CheckAllocation(allocErr, "stdColourMagnitude")
    stdColourMagnitude = NullValue  

    allocate(stdColourMagnitudeError(1 : nIndep, 1 : nFilt), stat=allocErr)
    call CheckAllocation(allocErr, "stdColourMagnitudeError")
    stdColourMagnitudeError = NullValue  
    noColours = 0 

    !***************************************************
    ! Calculate the standard magnitudes and colours
    ! NB- can only use U, B and V  
    !***************************************************
    allocate(standardMagnitude(1 : nIndep, 1 : nCol), stat=allocErr)
    call CheckAllocation(allocErr, "standardMagnitude")
    standardMagnitude = NullValue

    allocate(standardMagnitudeError(1 : nIndep, 1 : nCol), stat=allocErr)
    call CheckAllocation(allocErr, "standardMagnitudeError")
    standardMagnitudeError = NullValue    

    stdMagExist = .false.
    noColours = 0

    !*****************************************************************
    ! Calculate the colours in reverse order to preserve the priority
    !*****************************************************************
    do i = nCol1, 1, -1
      
       iColumn = Colour(i)%mag1
       jColumn = Colour(i)%mag2

       if (ANY(magexist(1 : nIndep, iColumn : iColumn)) .and.  &
            ANY(magexist(1 : nIndep, jColumn : jColumn))) then
          noColours = noColours + 1
          colourString(noColours) = Colour(i)%label
          colourErrorString(noColours) = Colour(i)%elabel

          stdMagExist(iColumn) = .true.
          stdMagExist(jColumn) = .true.
          
          do k = 1, nIndep
              if (mag_table(k, iColumn) .ne. NullValue .and. &
                  mag_table(k, jColumn) .ne. NullValue) then

                  !************************************************
                  ! Calculate the standard magnitudes and colour
                  !************************************************
                  call CalculateStandardColour(filt(iColumn), filt(jColumn), &
                       dble(mag_table(k, iColumn)), dble(mag_table(k, jColumn)), &
                       stdMagnitude1, stdMagnitude1Error, &
                       stdMagnitude2, stdMagnitude2Error, &
                       stdColour, stdColourError)
                  standardMagnitude(k, iColumn) = stdMagnitude1
                  standardMagnitudeError(k, iColumn) = stdMagnitude1Error 
                  standardMagnitude(k, jColumn) = stdMagnitude2
                  standardMagnitudeError(k, jColumn) = stdMagnitude2Error
                  stdColourMagnitude(k, noColours) = stdColour
                  stdColourMagnitudeError(k, noColours) = stdColourError
    
              end if
          end do 

       end if
    end do

  
    deallocate(magexist, stat=allocErr)
    call CheckDeallocation(allocErr, "magexist")

    deallocate(mag_table,stat=allocErr)
    call CheckDeallocation(allocErr, "mag_table")
    
    deallocate(magerr_table,stat=allocErr)
    call CheckDeallocation(allocErr, "magerr_table")
   
    !******************************************************
    ! Add the standard magnitude and its error to the table
    !******************************************************
    do i = 1 , nFilt

       if (stdMagExist(i))  then

          label = trim(filtid(i)) // "_STD"
          call Add2Real32Column(tableSet, label, "magnitudes", nIndep, standardMagnitude, i, RASortIndex, nullValue)

          label = trim(filtid(i)) // "_ERR_STD"
          call Add2Real32Column(tableSet, label, "magnitudes", nIndep, standardMagnitudeError, i, RASortIndex, nullValue)
       end if
    end do
 
    !********************************************
    ! Add the standard colours to the table
    !********************************************
    do i = 1, noColours
        label =  trim(colourString(i)) // "_STD"
        call Add2Real32Column(tableSet, label, "magnitudes", nIndep, stdColourMagnitude, i, RASortIndex, nullValue)
        label = trim(ColourErrorString(i)) // "_STD"
        call Add2Real32Column(tableSet, label, "magnitudes", nIndep, stdColourMagnitudeError, i, RASortIndex, nullValue)
    end do    
  
    deallocate(magColour, stat = allocErr)
    call CheckDeallocation(allocErr, "magColour")
    deallocate(magColourError, stat = allocErr)
    call CheckDeallocation(allocErr, "magColourError")
 
    deallocate(standardMagnitude, stat = allocErr)
    call CheckDeallocation(allocErr, "standardMagnitude")
  
    deallocate(standardMagnitudeError, stat = allocErr)
    call CheckDeallocation(allocErr, "standardMagnitudeError")

    deallocate(stdColourMagnitude, stat=allocErr)
    call CheckDeallocation(allocErr, "stdColourMagnitude")
    
    deallocate(stdColourMagnitudeError, stat=allocErr)
    call CheckDeallocation(allocErr, "stdColourMagnitudeError")
   
    !*************************************
    ! Add the major-axes for each filter 
    !*************************************
    do j = 1 , noFilters
 
       i = filterIndex(j)
       label = trim(filtid(i)) // "_MAJ"       
       call Add2Real32Column(tableSet, label, "pixels", nIndep, majorTable, i, RASortIndex, nullValue)
    end do

    deallocate(majorTable, stat=allocErr)
    call CheckDeallocation(allocErr, "majorTable")

    !*************************************
    ! Add the minor-axes for each filter 
    !*************************************
    do j = 1 , noFilters

        i = filterIndex(j)
        label = trim(filtid(i)) // "_MIN"
        call Add2Real32Column(tableSet, label, "pixels", nIndep, minorTable, i, RASortIndex, nullValue)

    end do

    deallocate(minorTable, stat=allocErr)
    call CheckDeallocation(allocErr, "minorTable")

    !*************************************
    ! Add the flag QFLAG
    !*************************************
    do j = 1 , noFilters
 
        i = filterIndex(j)
        label = trim(filtid(i)) // "_QFLAG"
        call AddInt8Column(tableSet, label, "none", nIndep, RASortIndex, qflag)
    end do
   
    deallocate(qflag, stat = allocErr)
    call CheckDeallocation(allocErr, "qflag")
    
    !*************************************
    ! Add the flag CFLAG
    !*************************************
    do j = 1, noFilters

        i = filterIndex(j)
        label = trim(filtid(i)) // "_CFLAG"
        call AddInt8Column(tableSet, label, "none", nIndep, RASortIndex, cflag)
 
    end do
    deallocate(cflag,stat=allocErr)
    call CheckDeallocation(allocErr, "cflag")

    !*************************************
    ! Add the flag EFLAG
    !*************************************
    do j = 1, noFilters

        i = filterIndex(j)
        label = trim(filtid(i)) // "_EFLAG"
        call AddInt8Column(tableSet, label, "none", nIndep, RASortIndex, eflag)
    end do

    deallocate(eflag,stat=allocErr)
    call CheckDeallocation(allocErr, "eflag")
  
    call release(tableSet)
    call release(outDataSet)

    !**********************************************************
    ! now we add an source ID column to each of the osw source lists
    ! telling us which source went to which entry in the
    ! observation source list (used by omsumm)
    !**********************************************************

    nStart = 1
    allocate(index(1 : nIndep), stat = allocErr)
    call CheckAllocation(allocErr, "index")
  
    forall (i = 1 : nIndep) index(RASortIndex(i)) = i

    ndims(1) = 1
    do i = 1 , nosw
        oswDataSet = DataSet(oswFile(i), modify)
        tableSet = table(oswDataSet, 0)
   
        if (hasColumn(tableSet, "SRC_ID")) then 
            call warning("colExist", "SRC)ID column exists in table")
            pcolumn = column(tableSet, "SRC_ID", modify)
        else
            pcolumn = addColumn(tableSet, "SRC_ID", integer32, " ", ndims)
        endif

        sourceID => int32Data(pcolumn)
        do j = nStart, nStart + nrows(i) - 1
            sourceID(j - nStart) = index(iIndep(oswtoobs(j, 3)))

        enddo
        call release(pcolumn)
      
        call release(tableSet)
        call release(oswDataSet)
        nStart = nStart + nrows(i)
    enddo


    deallocate(index, stat = allocErr)
    call CheckDeAllocation(allocErr, "index")
  

    deallocate(oswfile, stat=AllocErr)
    call CheckDeallocation(allocErr, "oswfile")
 
    deallocate(RASortIndex, stat = allocErr)
    call CheckDeallocation(allocErr, "RASortIndex")


    !*********************************************
    ! AMK - Tidy up memory that has been allocated
    !**********************************************
    deallocate(nrows, stat=allocErr)
    call CheckDeallocation(allocErr, "nrows")
    
    deallocate(oswtoobs, stat=allocErr)
    call CheckDeallocation(allocErr, "oswtoobs")

    deallocate(iIndep, stat=allocErr)
    call CheckDeallocation(allocErr, "iIndep")
    
end subroutine omsrclistcomb
!*******************************************************
! ArcLength - returns the distance (arc secs) between 2
! sources with coords (ra1, dec1) & (ra2, dec2)
! Returns the distance in arc seconds.
!*******************************************************
real(kind=double) function ArcLength(ra1, dec1, ra2, dec2)
    real(kind=double)              ::  ra1,ra2
    real(kind=double)              ::  dec1, dec2
    real(kind=double)              :: ang
    real(kind=double), parameter   :: pi = 3.14159265358979323846_double
    real(kind=double), save :: constant =  3600.0_double * 180.0_double / pi

    ang = abs(ra2 - ra1)
    if (ang > pi) ang = ang - pi

    ArcLength = dacos(sin(dec1) * sin(dec2) + cos(dec1) &
                * cos(dec2) * cos(ang)) * constant 

end function ArcLength
!****************************************************
! CalculateStandardColour - Given the two filters, two
! instrumental magnitudes and colours, calculates the 
! standard colours. If the instrumental colours are outside
! the validity range the values are set to nullValue.
!*****************************************************
subroutine CalculateStandardColour(filterID1, filterID2, &
           instrumentalMagnitude1, instrumentalMagnitude2, &
           standardMagnitude1, standardMagnitude1Error, & 
           standardMagnitude2, standardMagnitude2Error, &
           standardColour, standardColourError)
    integer(kind=int32) :: filterID1, filterID2
    real(kind=double), intent(in) :: instrumentalMagnitude1, instrumentalMagnitude2             
    real(kind=double), intent(out) :: standardMagnitude1, standardMagnitude2                     
    real(kind=double), intent(out) :: standardColour
    real(kind=double), intent(out) :: standardColourError
    real(kind=double), intent(out) :: standardMagnitude1Error
    real(kind=double), intent(out) :: standardMagnitude2Error  
    real(kind=double)    :: magnitudeRange
    type(OmColorTransformatorT)     :: colourTrans
    type(IntervalT)                 :: colourValInt
    
    colourTrans = CAL_omGetColorTransformator(filterID1, filterID2)
    colourValInt = CAL_omColorTransValidityRange(colourTrans)  

    !*******************************************************
    ! Check that the instrumental magnitudes lie within
    ! the allowed range.
    !*******************************************************
    magnitudeRange = instrumentalMagnitude1 - instrumentalMagnitude2
    if (magnitudeRange .gt. colourValInt%lower .and. &
        magnitudeRange .lt. colourValInt%upper)  then    ! Range okay                       
       
        !****************************************************
        ! NB The first magnitude in the following call is
        ! transformed into the standard system
        !****************************************************
        standardMagnitude1 = instrumentalMagnitude1
        call CAL_omColorTransform(filterId1, filterID2, &
                  standardMagnitude1, instrumentalMagnitude2, &
                  standardColour,  standardMagnitude2, &
                  standardColourError, standardMagnitude2Error)
        standardMagnitude1Error = AddErrors(standardMagnitude2Error,  standardColourError) 

    end if 
                                           
end subroutine CalculateStandardColour
!***************************************************
! CheckAllocation
!***************************************************
subroutine CheckAllocation(status, string)
    integer, intent(in) ::  status
    character(len=*), intent(in) :: string
    character(len=100) :: comment
   
    if(status .ne. 0) then
       write(comment, '(a32, a20)') "Could not allocate memory for : ", string 
       write (errstr, *) comment
       call fatal("allocErr") 
    end if

end subroutine CheckAllocation
!***************************************************
! CheckDeallocation
!***************************************************
subroutine CheckDeallocation(status, string)
    integer, intent(in) ::  status
    character(len=*), intent(in) :: string
    character(len=100) :: comment
   
    if(status .ne. 0) then
       write(comment, '(a31, a20)') "Failed to release memory for : ", trim(string) 
       write (errstr, *) comment
       call warning("allocErr", comment) 
    end if

end subroutine CheckDeallocation
!******************************************
! CreateObsSourceList
! create the output observation source list
!******************************************
subroutine  CreateObsSourceList(noOswFiles, oswFile, outDataSet)
    integer, intent(in) :: noOswFiles
    type(DataSetT), intent(out) :: outDataSet
    character(len=*), intent(in), dimension(1 : noOswFiles) :: oswFile
    character(len=40) :: startTimeString, endTimeString
    type(DataSetT) :: oswDataSet
    integer i, k
    character(len=40) :: startObs, endObs

    outDataSet = DataSet(stringParameter("outset"), create)
  
    call setAttribute(outDataSet, "HDUCLASS", "OGIP", "Format conforms to OGIP/GSFC conventions")
    call setAttribute(outDataSet, "HDUCLAS1", "SRCLIST", "File contains a source list")
    call setAttribute(outDataSet, "FILTER", "n/a", "Not applicable")
    call setAttribute(outDataSet, "CONTENT", "SRCLIST", "File content")

    !************************************************
    ! Find the first start time and the last end time
    !************************************************
    oswDataSet = DataSet(oswFile(1), read)

    startTimeString = stringAttribute(oswDataSet, "DATE-OBS")
    endTimeString   = stringAttribute(oswDataSet, "DATE-END")
    
    call release(oswDataSet)

    do i = 2, noOswFiles
 
        oswDataSet = DataSet(oswFile(i), read)
        startObs = stringAttribute(oswDataSet, "DATE-OBS")
        endObs   = stringAttribute(oswDataSet, "DATE-END")
        k = CompareDates(startTimeString, startObs)
        
        if (k .lt. 0) startTimeString = startObs

        k = CompareDates(endTimeString, endObs)
        
        if (k .gt. 0) endTimeString = endObs
             
        call release(oswDataSet)  
    end do

    call setAttribute(outDataSet, "DATE-OBS", startTimeString, "Start time of observations")
    call setAttribute(outDataSet, "DATE-END", endTimeString, "End time of observations")
    
end subroutine CreateObsSourceList
!**********************************************************
! ReadInt8Column
!
!*********************************************************
subroutine ReadInt8Column(tableSet, heading, size, start, array)
    type(TableT), intent(in) :: tableSet
    character(len=*), intent(in) :: heading
    integer, intent(in) :: size, start
    integer(kind=int8), intent(inout), dimension(:) :: array
    integer(kind=int8), pointer :: i8col(:)
    type(ColumnT) :: pColumn
    integer i, k
    
    k = start
    pColumn = column(tableSet, heading, read)
    i8col => int8Data(pcolumn)
    do i = 1 , size
        k = k  + 1
        array(k) = i8col(i - 1)
    enddo
    call release(pColumn)

end subroutine ReadInt8Column
!**********************************************************
! ReadInt32Column
!
!*********************************************************
subroutine ReadInt32Column(tableSet, heading, size, start, array)
    type(TableT), intent(in) :: tableSet
    character(len=*), intent(in) :: heading
    integer, intent(in) :: size, start
    integer(kind=int32), intent(inout), dimension(:) :: array
    integer(kind=int32), pointer :: i32col(:)
    type(ColumnT) :: pColumn
    integer i, k
    
    k = start
    pColumn = column(tableSet, heading, read)
    i32col => int32Data(pcolumn)
    do i = 1 , size
        k = k  + 1
        array(k) = i32col(i - 1)
    enddo
    call release(pColumn)

end subroutine ReadInt32Column
!**********************************************************
! ReadReal32Column
!
!*********************************************************
subroutine ReadReal32Column(tableSet, heading, size, start, array)
    type(TableT), intent(in) :: tableSet
    character(len=*), intent(in) :: heading
    integer, intent(in) :: size, start
    real(kind=single), intent(inout), dimension(:) :: array
    real(kind=single), pointer :: real32Col(:)
    type(ColumnT) :: pColumn
    integer i, k

    k = start
    pColumn = column(tableSet, heading, read)
    real32Col => real32Data(pcolumn)
    do i = 1 , size
        k = k + 1
        array(k) = real32Col(i - 1)
    enddo
    call release(pColumn)

end subroutine ReadReal32Column
!**********************************************************
! ReadReal64Column
!
!*********************************************************
subroutine ReadReal64Column(tableSet, heading, size, start, array)
    type(TableT), intent(in) :: tableSet
    character(len=*), intent(in) :: heading
    integer, intent(in) :: size, start
    real(kind=double), intent(inout), dimension(:) :: array
    real(kind=double), pointer :: real64Col(:)
    type(ColumnT) :: pColumn
    integer i, k
    
    k = start
    pColumn = column(tableSet, heading, read)
    real64Col => real64Data(pcolumn)
    do i = 1 , size
        k = k  + 1
        array(k) = real64Col(i - 1)
    end do
    call release(pColumn)

end subroutine ReadReal64Column
!***************************************************
! AddReal32Cololumn
!***************************************************
subroutine AddReal32Column(tableSet, heading, comment, &
        size, array, indexOrder, nullValue)
    type(TableT), intent(in) :: tableSet
    character(len=*), intent(in) :: heading, comment
    integer, intent(in) :: size, indexOrder(1 : size)
    real(kind=single), intent(in), dimension(:) :: array
    real(kind=single), intent(in) :: nullValue
    real(kind=single), pointer :: real32Col(:)
    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1) = (/1/)
 
    pcolumn = addColumn(tableSet, heading, real32, comment, ndims)
    real32Col => real32Data(pcolumn)

    do i = 1 , size
        if (array(indexOrder(i)) .eq. nullValue) then
            call setToNull(pcolumn, i - 1)
        else
            real32Col(i - 1) = array(indexOrder(i))
        end if
  
    enddo
    call release(pcolumn)

end subroutine AddReal32Column
!***************************************************
! Add2Real32Column
! For 2-dimensional arrays
!***************************************************
subroutine Add2Real32Column(tableSet, heading, comment, &
        size, array, index, indexOrder, nullValue)
    type(TableT), intent(in) :: tableSet
    character(len=*), intent(in) :: heading, comment
    integer, intent(in) :: size, index, indexOrder(1 : size)
    real(kind=single), intent(in), dimension(:, :) :: array
    real(kind=single), intent(in) :: nullValue
    real(kind=single), pointer :: real32Col(:)
    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1) = (/1/)
 
    pcolumn = addColumn(tableSet, heading, real32, comment, ndims)
    real32Col => real32Data(pcolumn)
  
    do i = 1 , size
        if (array(indexOrder(i), index) .eq. nullValue) then
            call setToNull(pcolumn, i - 1)
        else
            real32Col(i - 1) = array(indexOrder(i), index)
        end if
  
    enddo
    call release(pcolumn)

end subroutine Add2Real32Column
!***************************************************
! AddReal64Column
!***************************************************
subroutine AddReal64Column(tableSet, heading, comment, &
        size, array, indexOrder, nullValue)
    type(TableT), intent(in) :: tableSet
    character(len=*), intent(in) :: heading, comment
    integer, intent(in) :: size, indexOrder(1 : size)

    real(kind=double), intent(in), dimension(1 : size) :: array
    real(kind=single), intent(in) :: nullValue
    real(kind=double), pointer :: real64Col(:)
    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1) = (/1/)

    pcolumn = addColumn(tableSet, heading, real64, comment, ndims)
    real64Col => real64Data(pcolumn)

    do i = 1 , size
        if (array(indexOrder(i)) .eq. nullValue) then
            call setToNull(pcolumn, i - 1)
        else
            real64Col(i - 1) = array(indexOrder(i))
        end if
  
    enddo
    call release(pcolumn)

end subroutine AddReal64Column
!***************************************************
! AddInt8Cololumn
!***************************************************
subroutine AddInt8Column(tableSet, heading, comment, &
        size, indexOrder, array)
    type(TableT), intent(in) :: tableSet
    character(len=*), intent(in) :: heading, comment
    integer, intent(in) :: size, indexOrder(1 : size)
    integer(kind=int8), intent(in), dimension(:) :: array
    integer(kind=int8), pointer :: int8Col(:)
    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1) = (/1/)

    pcolumn = addColumn(tableSet, heading, integer8, comment, ndims)
    int8Col => int8Data(pcolumn)

    forall (i = 1 : size) int8Col(i - 1) = array(indexOrder(i))
  
    call release(pcolumn)

end subroutine AddInt8Column

!***************************************************
! AddInt32Cololumn
!***************************************************
subroutine AddInt32Column(tableSet, heading, comment, &
        size, indexOrder, array)
    type(TableT), intent(in) :: tableSet
    character(len=*), intent(in) :: heading, comment
    integer, intent(in) :: size, indexOrder(1 : size)
    integer(kind=int32), intent(in), dimension(:) :: array
    integer(kind=int32), pointer :: int32Col(:)
    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1) = (/1/)

    pcolumn = addColumn(tableSet, heading, integer32, comment, ndims)
  
    int32Col => int32Data(pcolumn)

    forall(i = 1: size) int32Col(i - 1) = array(indexOrder(i))

    call release(pcolumn)

end subroutine AddInt32Column
!*******************************************************
! AddErrors
!
!*******************************************************
real(kind=double) function AddErrors(error1, error2)
    real(kind=double), intent(in) :: error1, error2
 
    AddErrors = dsqrt(error1 * error1  + error2 * error2)

end function AddErrors
!******************************************************
! subroutine SetUpColourTable
! Initialises the array with the possible colours
!******************************************************
subroutine SetUpColourTable(Colour)

    type(ColourT), intent(out), dimension(:) :: Colour
   
    !***************** 
    ! These are OPT-OPT
    !*****************
    Colour(1) = ColourT(5, 6, "U_B", "U_B_ERR")
    Colour(2) = ColourT(6, 7, "B_V", "B_V_ERR")
    Colour(3) = ColourT(5, 7, "U_V", "U_V_ERR")

    !****************
    ! These are UV-UV
    !****************
    Colour(4) = ColourT(2, 4, "UVW2_UVW1", "UVW2_UVW1_ERR")
    Colour(5) = ColourT(2, 3, "UVW2_UVM2", "UVW2_UVM2_ERR")
    Colour(6) = ColourT(3, 4, "UVM2_UVW1", "UVM2_UVW1_ERR")

    !*****************
    ! These are UV-OPT
    !*****************
    Colour(7)  = ColourT(2, 5, "UVW2_U", "UVW2_U_ERR")
    Colour(8)  = ColourT(2, 6, "UVW2_B", "UVW2_B_ERR")
    Colour(9)  = ColourT(2, 7, "UVW2_V", "UVW2_W_ERR")
    Colour(10) = ColourT(3, 5, "UVM2_U", "UVM2_U_ERR")
    Colour(11) = ColourT(3, 6, "UVM2_B", "UVM2_B_ERR")
    Colour(12) = ColourT(3, 7, "UVM2_V", "UVM2_V_ERR")
    Colour(13) = ColourT(4, 5, "UVW1_U", "UVW1_U_ERR")
    Colour(14) = ColourT(4, 6, "UVW1_B", "UVW1_B_ERR")
    Colour(15) = ColourT(4, 7, "UVW1_V", "UVW1_V_ERR")

end subroutine SetUpColourTable
!*******************************************************
! CompareDates - Compares the first date with the second.
! Returns -1 if the second is less than the first,
! 0 if they are equal and 1 if greater.
!*******************************************************
integer function CompareDates(dateString1, dateString2)
    character(len = *), intent(in) :: dateString1, dateString2
    integer :: time1(6), time2(6)        ! Stores the time from the date string
    character(len = 1) char    ! Used for date and time extraction 
    integer i

    read(dateString1, '(i4, a1, i2, a1, i2, a1, i2, a1, i2, a1, i2)') &
         time1(1), char, time1(2), char, time1(3), char, time1(4), &
         char, time1(5), char, time1(6)

    read(dateString2, '(i4, a1, i2, a1, i2, a1, i2, a1, i2, a1, i2)') &
         time2(1), char, time2(2), char, time2(3), char, time2(4), &
         char, time2(5), char, time2(6)
  
    do i = 1, 6

        CompareDates = time2(i) - time1(i)
        if (CompareDates .ne. 0) return
 
    end do

    CompareDates = 0 ! Must be equal

end function CompareDates
!**********************************
! ComputeAverage
! Calculates the weighted average and
! standard deviation
!**********************************
subroutine ComputeAverage(data, error, size, mean, standardDeviation)
    integer, intent(in) :: size        
    real(kind=double), intent(in), dimension(1 : size) :: data, error    
    real(kind=double), intent(out) :: mean, standardDeviation
    real(kind=double) :: weight = 1.0_double
    real(kind=double) :: sum1, sum2, sum3
    real(kind=double) :: d, difference
    logical :: useWeights = .true.
    integer i


    if (size .eq. 1) then
        mean = data(1)
        standardDeviation = error(1)
        return
    end if

    do i = 1, size
        if (error(i) .le. 0.0) then
            useWeights = .false.
            exit
        end if
    end do 

    sum1 = 0.0d0
    sum2 = 0.0d0
    sum3 = 0.0d0
  
    do i = 1, size
        if (useWeights) weight = 1.0d0 / (error(i) * error(i))
        sum1 = sum1 + weight
        sum2 = sum2 + data(i) * weight
        sum3 = sum3 + weight * weight
    end do

    mean = sum2 / sum1  ! The mean

    sum2 = 0.0d0 
    do i = 1, size
        if (useWeights) weight = 1.0d0 / (error(i) * error(i))
        difference = data(i) - mean
        sum2 = sum2 + weight * difference * difference
    end do
    d = sum1 - sum3 / sum1

    standardDeviation = dsqrt(sum2 / d) ! The standard deviation

end subroutine ComputeAverage
!************************************************************
! Subroutine HeapSort
! Sorts  one-dimensional array(1.. size) int ascending order.
! An index of the sorted order relative to the original
! order is returned.
!******************** 

subroutine HeapSort(array, index, size)
    integer, intent(in) :: size
    double precision, intent(in), dimension(1 : size) :: array
    integer, intent(out), dimension(1 : size) :: index

    integer :: i, ir, j, l, irra
    double precision :: rra
  
    if (size < 2) return


    l = size / 2 + 1
    ir = size   
    do i = 1, size
        index(i) = i
       
    end do   

    do 
        if (l > 1) then
            l = l - 1
            rra = array(index(l))   
            irra = index(l)    
        else
            rra = array(index(ir))        
            irra = index(ir)
          
            index(ir) = index(1)
            ir = ir - 1
            
            if (ir <= 1) then
             
                index(1) = irra
                exit
            end if
        end if

        i = l
        j = l + l
        do while(j <= ir)
       
            if (j < ir) then
                if(array(index(j)) < array(index(j + 1))) j = j + 1
            end if
            if (rra < array(index(j))) then
               
                index(i) = index(j)
                i = j
                j = j + j
            else
                exit
            end if
            
        end do   
          
        index(i) = irra
    end do
  
end subroutine HeapSort


end module omsrclistcomb_mod




















